--[[
string-test.lua

testing the string library
]]

LUA_PATH = "?.lua"

require "common"

printer = console.printer()
print = printer.print

-- uncomment the following line to have all output in the clipboard (only if the script terminates normally)
--want_text_in_clipboard = true


-- the main program
function main()
    local auml = "\195\164"
    local ouml = "\195\182"
    local uuml = "\195\188"
    local Auml = "\195\132"
    local Ouml = "\195\150"
    local Uuml = "\195\156"
    local szlig = "\195\159"

    -- announce ourselves
    console.popup()
    print("now running: string-test.lua")
    
    --local ansi = "%s"
    local str_utf8 = auml..ouml..uuml.."%s"..Auml..Ouml..Uuml..szlig
    
    print("test string: ", quote(str_utf8))
    
    -- length of strings
    -- utf8.length(s) - assumes s is a utf8 encoded string. returns the number of characters in s.
    -- string.length(s) - returns the number of bytes in the string s.
    print("utf8.length: ", utf8.length(str_utf8))
    print("string.length: ", string.length(str_utf8))
    
    -- finding substrings
    -- utf8.find(s1, s2, from = 1) - for strings s1 and s2, if s2 is contained in s1, returns the position of s2 in s1,
    --     otherwise returns nil. starts searching at character at index from. the position is returned as the 
    --     indices of the character in s1 where s2 starts and ends.
    -- string.find(s1, s2, from = 1) - like find but all indices refer to bytes instead of characters
    --
    -- the original lua string functions find, gfind, and gsub are available as ansifind, ansigfind, and ansigsub
    local from, to = utf8.find(str_utf8,uuml)
    print("position of ",uuml," in test string (using utf8.find): ", from and (from .. " -> " .. to) or "not found")
    from, to = string.find(str_utf8,uuml)
    print("position of ",uuml," in test string (using string.find): ", from and (from .. " -> " .. to) or "not found")
    
    -- extracting substrings
    -- utf8.sub(s, start, end = -1) - returns the substring from character start to character end (inclusive).
    -- string.sub(s, start, end = -1) - the lua sub function. like substring, but indices refer to bytes.
    print("testing substring extraction...")
    print("utf8.sub from character 2 to 4: ", quote(utf8.sub(str_utf8, 2, 4)) or "failed")
    print("string.sub from byte 3 to 7: ", quote(string.sub(str_utf8, 3, 7)) or "failed")
    
    -- character set conversion
    -- utf8.from_ansi(s) - assumes s is an ansi encoded string. returns utf8 version of s.
    -- utf8.to_ansi(s) - assumes s is a utf8 encoded string. returns ansi version of s.
    --
    -- conversion to and from wide string is not supported by now. however, decode and encode may help.
    print("testing character set conversion...")
    local str_ansi = utf8.to_ansi(str_utf8)
    print("to ansi and back: ", str_ansi and quote(utf8.from_ansi(str_ansi)) or "failed")
    
    -- case conversion
    -- lower(s) - returns upper case version of string s
    -- upper(s) - returns lower case version of string s
    print("testing case conversion...")
    print("lower case: ", quote(utf8.lower(str_utf8)))
    print("upper case: ", quote(utf8.upper(str_utf8)))
    
    -- formatting
    -- format(spec, ...) - options in available in spec are the same as for printf in C/C++. exceptions are
    --     *, l, L, n, p, h which are not supported. there is an extra option q which formats a string in a way
    --     that it can be read back into lua. string values to be formatted with %s cannot contain embedded zeros.
    print("testing formatting...")
    print("applying format to test string: ", quote(string.format(str_utf8, str_utf8)) or "failed")
    
    -- repeating strings
    -- rep(s, n) - returns a string that is the concatenation of n copies of the string s.
    --      attention: do not confuse with repeat, which is a keyword.
    print("testing repeat...")
    print("five times fb2k: ", quote(string.rep("fb2k", 5)) or "failed")

    -- utf8.decode and utf8.encode
    -- utf8.decode(s, start, end = start, forcetable = false) - if end == start, decode returns the character 
    --     code of the character at start as a number. if end > start or forcetable is true, decode returns
    --     the codes of all characters from start to end in a table. both start and end are byte indices.
    --     if end points into the middle of a character, that character is still returned.
    -- utf8.encode(...) - the inverse operation to decode. takes a list of numbers or a table containing numbers,
    --     and returns the string with the characters responding to the given character codes.
    print("testing utf8.decode and utf8.encode...")
    local code = utf8.decode(str_utf8,1)
    local char = code and (code > 0) and quote(utf8.encode(code)) or "failed"
    print("decode of character at start: ", code or "failed")
    print("encode of previous value: ", code and quote(utf8.encode(code)) or "failed")
    local t = utf8.decode(str_utf8,1,-1)
    print("decode of whole string: ", t and quote(t) or "failed")
    print("encode of previous value: ", t and quote(utf8.encode(t)) or "failed")
    t = string.decode(str_utf8, 1, 1, true)
    print("decode of character at start, forcing table: ", t and quote(t) or "failed")
    
    -- string.decode and string.encode
    -- are like decode and encode, but they operate on bytes rather than characters.
    print("testing string.encode and string.decode...")
    code = string.decode("a",1)
    print("decode of character a: ", code or "failed")
    print("encode of previous value: ", code and quote(string.encode(code)) or "failed")
    t = string.decode(str_utf8,1,-1)
    print("decode of whole string: ", t and quote(t) or "failed")
    print("encode of previous value: ", t and quote(string.encode(t)) or "failed")
    t = string.decode(str_utf8, 1, 1, true)
    print("decode of character at start, forcing table: ", t and quote(t) or "failed")
    
    -- copy all output to clipboard
    if want_text_in_clipboard then
        printer.copy()
    end
end

main()
