--[[*
Generates report on database contents.

@author foosion
@requires foo_lua 0.2
@command Components/Reports/Database statistics
]]

assert(_COMMANDMODE == 'main menu',
	'this script needs to be run from foo_lua as a main menu command')
	
assert(fb2k and fb2k.metadb or loadmodule("fb2k.metadb"))

prefs = {}
local defaults = {}

-- get a unique value
IGNORE = {}

defaults.show_dialog = 'no'
defaults.generate_report = 'yes'
defaults.choose_template_file = 'yes'
defaults.template_file = nil
defaults.choose_output_file = 'yes'
defaults.output_file = nil

defaults.protocol_map = {}
defaults.extension_map = {}

local function set_defaults()
	for k, v in pairs(defaults) do
		if not prefs[k] then
			prefs[k] = v
		end
	end
end

local function loadprefs()
	dofile(_SCRIPTDIR..'dbstats-prefs.lua')
end

-- read in preferences
local res, err = pcall(loadprefs)
if not res then
	fb2k.console.error('error while reading preferences:')
	fb2k.console.error(err)
	fb2k.console.info('using default preferences')
end

set_defaults()

-----------------------------
-- builtin template
-----------------------------

local builtin_template = [[Tracks by codec [count]
<<for _, codec in ipairs(codecs_by_count) do>>
<(codec.name)>: <(codec.count)> (<(codec.percent_count)>%)
<<end>>
total: <(total_count)> (100.0%)

Tracks by codec [duration]
<<for _, codec in ipairs(codecs_by_duration) do>>
<(codec.name)>: <(format_duration(codec.duration))> (<(codec.percent_duration)>%)
<<end>>
total: <(format_duration(total_duration))> (100.0%)
]]

-----------------------------
-- code
-----------------------------


-----------------------------
-- buffer class
-----------------------------

buffer = {}

function buffer.create()
	local t = {}
	t.data = {}
	function t:output(...)
		for _,v in ipairs(arg) do
			table.insert(self.data, v)
		end
	end
	function t:reset()
		self.data = {}
	end
	function t:tostring()
		return table.concat(self.data)
	end
	return t
end

-----------------------------
-- LHTML module
-----------------------------

lhtml = {}

function lhtml.parse2(buf, s)
	while s ~= '' do
		local b, e, pre, code = string.find(s, '^(.-)<%((.-)%)>')
		if b then
			buf:output(string.format('write_raw(%q);', pre))
			buf:output(string.format('write(%s);', code))
			s = string.sub(s, e+1)
		else
			buf:output(string.format('write_raw(%q);', s))
			s = ''
		end
	end
end

function lhtml.parse(s)
	local buf = buffer.create()
	while s ~= '' do
		local b, e, pre, code, nl = string.find(s,'^(.-)<<(.-)>>(\n?)')
		if b then
			lhtml.parse2(buf, pre)
			buf:output(code,' ',nl)
			s = string.sub(s,e+1)
		else
			lhtml.parse2(buf, s)
			s = ''
		end
	end
	return buf:tostring()
end

function lhtml.compile(s)
	local pre =  [[
local buf = buffer.create()
	
local function write_raw(...)
	buf:output(table.concat(arg))
end

local function write(...)
	local xlate_table = {
		['<'] = '&lt;',
		['>'] = '&gt;',
		['&'] = '&amp;',
	}
	
	local function xlate(s)
		return xlate_table[s] or ''
	end
	
	for _, s in ipairs(arg) do
		local s2 = string.gsub(tostring(s), '[<>&]', xlate)
		buf:output(s2)
	end
end
]]
	
	local post = [[
return buf:tostring()
]]

	local s2 = lhtml.parse(s)
	
	return loadstring(pre..s2..post, 'LHTML')
end

-----------------------------
-- main
-----------------------------

-- global values that the HTML template can access
data = nil
total_count = nil
total_duration = nil
max_count = nil
max_duration = nil
codecs = nil
codecs_by_count = nil
codecs_by_duration = nil

local UNKNOWN = "unknown"

-- get codec from metadb_handle
function get_codec(x)
	local info = x:query()
	if info then
		local codec = info:info_get("codec")
		if codec then
			return codec
		end
		local path, _ = info:get_location() -- ignore subsong index
		local _, _, protocol = string.find(path, '^(.-)://')
		if protocol then
			local codec = prefs.protocol_map[utf8.lower(protocol)]
			if codec then
				return codec
			end
		end
		local _, _, extension = string.find(path, '.*%.(.-)$')
		if extension then
			local codec = prefs.extension_map[utf8.lower(extension)]
			if codec then
				return codec
			end
		end
	end
	return UNKNOWN
end

function format_duration(x)
	local res = ''
	local s = math.floor(x)
	local m = math.floor(s / 60)
	local h = math.floor(m / 60)
	s = math.mod(s, 60)
	m = math.mod(m, 60)
	local d = math.floor(h / 24)
	h = math.mod(h, 24)
	res = string.format('%02d:%02d:%02d', h, m, s)
	
	if d > 0 then
		if d >= 7 then
			local wk = math.floor(d / 7)
			d = math.mod(d, 7)
			res = wk..'wk '..d..'d '..res
		else
			res = d..'d '..res
		end
	end
	return res
end

do
	local _info = fb2k.console.info
	function fb2k.console.info(...)
		_info(table.concat(arg),_SCRIPTFILE)
	end
	
	local _warning = fb2k.console.warning
	function fb2k.console.warning(...)
		_warning(table.concat(arg),_SCRIPTFILE)
	end

	local _error = fb2k.console.error
	function fb2k.console.error(...)
		_error(table.concat(arg),_SCRIPTFILE)
	end
end

-- UI helper
function dialog.ask(question,option)
	if option == 'yes' then
		return true
	elseif option == 'no' then
		return false
	elseif option == 'ask' then
		local d = dialog
		local answer = d.messagebox(question,'Database stats',d.MB_YESNO,d.MB_ICONQUESTION)
		return answer == d.IDYES
	else
		error('invalid argument')
	end
end

function dialog.notify(msg)
	local d = dialog
	d.messagebox(msg,'Database stats',d.MB_OK,d.MB_ICONINFORMATION)
end

function dialog.error(msg)
	local d = dialog
	d.messagebox(msg,'Database stats',d.MB_OK,d.MB_ICONERROR)
end

function categorize(data,getkey)
	local t = {}
	
	for i,v in ipairs(data) do
		local key = getkey(v) or UNKNOWN
		if key ~= IGNORE then
			if not t[key] then t[key] = {} end
			table.insert(t[key],v)
		end
	end
	
	local by_key = {}
	
	for k,v in pairs(t) do
		local cat = {name=k,count=table.getn(v),data=v}
		table.insert(by_key,cat)
	end
	
	return by_key
end

function format_percent(x, y)
	local p = x
	if y then
		if y == 0 then
			p = 0
		else
			p = x / y
		end
	end
	p = p * 100
	return math.floor(p).."."..math.mod(math.floor(p*10),10)
end

function prepare_data_bycodec(data)
	local codecs = categorize(data,get_codec)
	local unknown_codec = nil
	local max = codecs[1] and codecs[1].count or 0

	local t = {}

	total_count = 0
	max_count = 0
	total_duration = 0
	max_duration = 0
	
	for i, v in ipairs(codecs) do
		v.id = "id"..i
		if v.name == UNKNOWN then
			unknown_codec = i
		end
		table.insert(t,v)

		
		local duration = 0
		for _, track in ipairs(v.data) do
			local len = track:get_length()
			if len > 0 then
				duration = duration + len
			end
		end
		v.duration = duration
		
		total_count = total_count + v.count
		total_duration = total_duration + v.duration
		
		if v.count > max_count then
			max_count = v.count
		end
		if v.duration > max_duration then
			max_duration = v.duration
		end
	end
	
	for _, v in codecs do
		v.percent_count = format_percent(v.count, total_count)
		v.percent2_count = format_percent(v.count, max_count)
		v.percent_duration =  format_percent(v.duration, total_duration)
		v.percent2_duration =  format_percent(v.duration, max_duration)
	end
	
	return t, unknown_codec
end

function main(data)
	-----------------------------
	-- prepare data
	-----------------------------
	
	local show_dialog = prefs.show_dialog == 'yes'
	local generate_report = dialog.ask('Do want to generate an HTML file?', prefs.generate_report)
	
	local unknown_codec
	
	if show_dialog or generate_report then
		fb2k.console.info('preparing data')
	
		codecs, unknown_codec = prepare_data_bycodec(data)
	
		codecs_by_count = {}
		codecs_by_duration = {}
		for i, v in ipairs(codecs) do
			if i ~= unknown_codec then
				table.insert(codecs_by_count, v)
				table.insert(codecs_by_duration, v)
			end
		end
		
		table.sort(codecs_by_count,function(x,y) return x.count > y.count end)
		table.sort(codecs_by_duration,function(x,y) return x.duration > y.duration end)
		
		if unknown_codec then
			table.insert(codecs_by_count, codecs[unknown_codec])
			table.insert(codecs_by_duration, codecs[unknown_codec])
		end
	end
	
	-------------------------
	-- show dialog
	-------------------------
	
	if show_dialog then
		local buf = buffer.create()
		
		buf:output(string.format("total: %d (100.0%%)\n",total_count))
		
		for k,v in pairs(codecs_by_count) do
			buf:output(string.format("%s: %d (%s%%)\n",v.name,v.count,v.percent_count))
		end
		
		local text = buf:tostring()
		
		dialog.notify(text)
	end
	
	-------------------------
	-- generate HTML
	-------------------------
	
	if generate_report then
		local template
		if prefs.choose_template_file == 'builtin' then
			fb2k.console.info('using builtin template')
			template = builtin_template
			if not template then
				fb2k.console.error('no builtin template available')
			end
		else
			local template_file = prefs.template_file
			if prefs.choose_template_file == 'yes' or not prefs.template_file then
				template_file = dialog.choose_file('LHTML Files|*.lhtml|All Files|*.*', --ext_mask
					1, -- def_ext_mask
					'lhtml', -- def_ext
					'Load template file', -- title
					template_file, -- path
					false -- save
					)
			end
			if template_file then
				fb2k.console.info('opening template file for reading:')
				fb2k.console.info('location: "'..template_file..'"')
				local s = file.open(template_file, stream.MODE_READ)
				if s then
					local len = s:get_length()
					template = s:read(len)
					s:release()
				else
					fb2k.console.error('could not open file')
				end
			end
		end
		if template then
			fb2k.console.info('compiling template')
			local f, err = lhtml.compile(template)
			
			if f then
				fb2k.console.info('creating document from template')
				text = f()
			else
				error(err)
			end
			
			local output_file = prefs.output_file
			if prefs.choose_output_file == 'yes' or not prefs.output_file then
				output_file = dialog.choose_file('HTML Files|*.html|Text Files|*.txt|All Files|*.*', --ext_mask
					1, -- def_ext_mask
					'html', -- def_ext
					'Save file', -- title
					output_file, -- path
					true -- save
					)
			end
			if output_file then
				fb2k.console.info('opening document file for writing:')
				fb2k.console.info('location: "'..output_file..'"')
				local s = file.open(output_file, stream.MODE_WRITE_NEW)
				if s then
					s:write(text)
					s:release()
				else
					fb2k.console.error('could not open file')
				end
			end
		end
	end
end

data = fb2k.metadb.get_all_entries()
main(data)
