require "common"

-- Groups items from a list.
-- A group is a list of items, the relative order of items
-- in a group is the same as in the original list.
-- Items that have the same key as determined by get_key(item)
-- will be in the same group, only if they are adjacent in
-- the original list.
-- Example:
-- list = {1, 2, 3, 4, 5, 6}
-- function get_key(item) return math.mod(n, 3) == 0 and 1 or 0 end
-- group_locally(get_key, list) => {{1, 2}, {3}, {4, 5}, {6}}
function group_locally(get_key, list)
	local result = {}
	local groups = {}
	
	for _, item in list do
		-- get key
		local key = get_key(item)
		-- get group
		local group = key and groups[key]
		if not group then
			-- update groups
			group = {}
			groups = {[key] = group}
			-- add group to result
			table.insert(result, group)
		end
		-- add item to group
		table.insert(group, item)
	end
	
	return result
end

-- Groups items from a list.
-- A group is a list of items, the relative order of items
-- in a group is the same as in the original list.
-- Items that have the same key as determined by get_key(item)
-- will be in the same group.
-- Example:
-- list = {1, 2, 3, 4, 5, 6}
-- function get_key(item) return math.mod(n, 3) == 0 and 1 or 0 end
-- group_globally(get_key, list) => {{1, 2, 4, 5}, {3, 6}}
function group_globally(get_key, list)
	local result = {}
	local groups = {}
	
	for _, item in list do
		-- get key
		local key = get_key(item)
		-- get group
		local group = key and groups[key]
		if not group then
			-- update groups
			group = {}
			groups[key] = group
			-- add group to result
			table.insert(result, group)
		end
		-- add item to group
		table.insert(group, item)
	end
	
	return result
end

-- Flattens a list of groups.
-- Example:
-- list = {{1, 2, 4, 5}, {3, 6}}
-- ungroup(list) => {1, 2, 4, 5, 3, 6}
function ungroup(list)
	local result = {}
	
	for i,group in list do
		for j,v in group do
			table.insert(result, v)
		end
	end
	
	return result
end

local function self_test()
	local list = {1, 2, 3, 4, 5, 6}
	local function get_key(n)
		return math.mod(n, 3) == 0 and 1 or 0
	end
	
	local function test(name, expected, result)
		print(name,"=> ", quote(result), equiv(expected, result) and "(passed)" or "(failed)")
	end
	
	print("list = ", quote(list))
	local list2 = group_locally(get_key,list)
	test("group locally", list2, {{1, 2}, {3}, {4, 5}, {6}})
	local list3 = group_globally(get_key,list)
	test("group globally", list3, {{1, 2, 4, 5}, {3, 6}})
	local list4 = ungroup(list3)
	test("ungroup", list4, {1, 2, 4, 5, 3, 6})
end

self_test()
